
/*****************************************************************************
	hCount for use with the new Tilted Time Windows with Subtraction
	01-NOV-2007
*****************************************************************************/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#define MAXINT INT_MAX
#include "hcsub.h"

extern long N;		// current transaction number
extern long BSIZE;	// batch size

/***************************************************************************** 
hc_type *hc_init(float rho, float epsilon, long M)
INPUTS:
	rho - hCount parameter
	epsilon - hCount parameter
	M - largest item in the stream

OUTPUT:
	Pointer to initialized hc_type struct

Creates an instance of the hc_type struct and initializes everything
*****************************************************************************/
hc_type *hc_init(float rho, float epsilon, long M)
{
    float e;			// min x/ln(x) and x > 1  --> evaluates to the base of the natural logarithm
    int R_RANGE;		// range for random numbers a_i and b_i
   int i;
    TS_TYPE *aptr;

    hc_type *result = calloc(1, sizeof(hc_type));
    if (result == NULL) {
		printf("\ncalloc error");
		exit(100);
    }

    result->N = 0;			// total transactions
    result->M = M; 
    e = 2.718281828;
    //result->P = 1048583;	// big prime number
	//result->P = 1021;	
	result->P = 65539;
	//result->P = 16411;

    R_RANGE = result->M;
    result->m =  e / epsilon;		// length of hash table 
    result->h = (int) (log(-result->M / log(rho)));		// width of hash table
    
	// Allocate memory for the counters S
	aptr = (TS_TYPE *) malloc(result->m * result->h * sizeof(TS_TYPE));
    if (aptr == NULL) {
		printf("\nmalloc() error.");
		exit(100);
    }
    result->S = (TS_TYPE **) malloc(result->m *sizeof(TS_TYPE));
    if (result->S == NULL) {
		printf("\nmalloc() error.");
		exit(200);
    }
    for (i = 0; i < result->m; i++) {
		result->S[i] = aptr + i * result->h;
    }
    
	// Set all counters to 0
	memset((void *) aptr, 0, result->m * result->h * sizeof(TS_TYPE) / sizeof(char));
    
	// generate h pairs of random numbers a_i and b_i
	result->a = (int *) malloc(result->h * sizeof(int));
    if (result->a == NULL) {
		printf("\nmalloc() error.");
		exit(300);
    }
    result->b = (int *) malloc(result->h * sizeof(int));
    if (result->b == NULL) {
		printf("\nmalloc() error.");
		exit(400);
    }
    for (i = 0; i < result->h; i++) {
		result->a[i] = random(0, R_RANGE);
		result->b[i] = random(0, R_RANGE);
    }

	return (result);
}

/*****************************************************************************
void hc_update(hc_type * hc, int k, int ttype)
INPUTS:
	hc - pointer to the hCount struct
	k - the item in the stream
	ttype - transaction type: 1 for INSERT, -1 for DELETE

OUTPUT:
	NIL

Updates the hCount table with new transaction. 
*****************************************************************************/
void hc_update(hc_type * hc, int k, int ttype)
{
    int pos, j;
	
	hc->N++;
	N++;

    for (j = 0; j < hc->h; j++) {
		pos = ((((long long) hc->a[j] * k + hc->b[j]) % hc->P) % hc->m);

		if (ttype == INSERT) {
			hc->S[pos][j].c[0] = hc->S[pos][j].c[0] + 1;
		} else {
			hc->S[pos][j].c[0] = hc->S[pos][j].c[0] - 1;
		}
    }

	// Shift the Tilted Time Windows if necessary
	if ((hc->N - 1)% BSIZE == 0) {
        hc_shift(hc);
    }
}

/*****************************************************************************
void hc_shift(hc_type *hc)
INPUTS:
	hc - pointer to the hCount struct
	
OUTPUT:
	NIL

Shifts the Tilted Time Windows associated with each counter
*****************************************************************************/

#define ORIGVERSION 1

#ifdef ORIGVERSION
void hc_shift(hc_type *hc)
{
	int i, j, k;
	int index, powerValue, threshold;

	for (i=0; i<hc->m; i++) {
		for (j=0; j<hc->h; j++) {
			index = hc->N/BSIZE;	
			powerValue = 0;
			for (powerValue = WINDOWS-2; powerValue >= 0; powerValue--) {
				threshold = (int)pow(2.0, (double)powerValue);
				if (hc->N > 0 && index % threshold == 0) {
					hc->S[i][j].c[powerValue + 1] = 0;	
					for(k = 0; k < powerValue + 1; k++)
						hc->S[i][j].c[powerValue + 1] += hc->S[i][j].c[k];
				}
			}
			hc->S[i][j].c[0] = 0;	
		}
	}
}
#else // these two versions are equivalent. the second is fastest because it cuts many loops on m and h if the threshold condition does not hold.
void hc_shift(hc_type *hc)
{
    int i, j, k;
    int index, powerValue, threshold;


    index = hc->N/BSIZE;
    powerValue = 0;
    for (powerValue = WINDOWS-2; powerValue >= 0; powerValue--) {
        threshold = (int)pow(2.0, (double)powerValue);
        if (hc->N > 0 && index % threshold == 0) {

            // printf("shift a N=%d powerValue=%d\n", hc->N, powerValue);

            for (i=0; i<hc->m; i++) {
                for (j=0; j<hc->h; j++) {

                    hc->S[i][j].c[powerValue + 1] = 0;
                    for(k = 0; k < powerValue + 1; k++) {
                        hc->S[i][j].c[powerValue + 1] += hc->S[i][j].c[k];
                    }
                }
            }

        }
    }

    for (i=0; i<hc->m; i++) {
        for (j=0; j<hc->h; j++) {
            hc->S[i][j].c[0] = 0;
        }
    }

}

#endif

/*****************************************************************************
long hc_query(hc_type *hc, int id, long item)	
INPUTS:
	hc - pointer to the hCount struct
	id - ID of the Tilted Time Window
	item - item to look for

OUTPUT:
	Count of "item" in window number "id"

Counts the number of occurences of "item" in window number "id" (This is the
core query part of the hCount algorithm)
*****************************************************************************/
long hc_query(hc_type *hc, int id, long item)	
{
	long min = MAXINT;
	int j, pos, c;

	for (j = 0; j < hc->h; j++) {
		pos = ((((long long) hc->a[j] * item + hc->b[j]) % hc->P) % hc->m);
		c = hc->S[pos][j].c[id];
		if (c < min) {
			min = c;
		}
	}
	
	return(min);
}


/*****************************************************************************
int hc_size(hc_type * hc)
INPUTS:
	hc - pointer to the hCount structure

OUTPUT:
	Space occupied by the hCount structure in bytes
*****************************************************************************/
int hc_size(hc_type * hc)
{
    int size;
    size = hc->m * hc->h * sizeof(TS_TYPE);
    return (size);
}


/*****************************************************************************
void hc_print(hc_type *hc, int id)
INPUTS:
	hc - pointer to the hCount structure
	id - ID of Tilted Time Window

OUTPUT:
	NIL

Prints relevant debug information about the hCount structure
*****************************************************************************/
void hc_print(hc_type *hc, int id)
{
    int i, j;

    for (i = 0; i < hc->m; i++) {
		printf("\n");
		for (j = 0; j < hc->h; j++)
			printf("%d ", hc->S[i][j].c[id]);
    }

    for (i = 0; i < hc->h; i++) {
		printf("\n%d, a=%d, b=%d", i, hc->a[i], hc->b[i]);
    }

    printf("\nm = %d", hc->m);
    printf("\nh = %d", hc->h);
    printf("\nP = %d", hc->P);
    printf("\nhc_Print -- END");

}

/*****************************************************************************
void hc_destroy(hc_type * hc)
INPUTS:
	hc - pointer to the hCount structure

OUTPUT:
	NIL

Frees up memory used by the hCount structure
*****************************************************************************/
void hc_destroy(hc_type * hc)
{
    free(hc->a);
    free(hc->b);
    free(hc->S);
}

